
// Based on Archive_cpp.template

/*****************************************************************************/
/* Copyright (C) 2003-2010, JOANNEUM RESEARCH Forschungs Ges.m.b.H.          */
/* Steyrergasse 17, A-8010 Graz, Austria.                                    */
/*                                                                           */
/*  This file is part of the JRS MPEG-7 Library.                             */
/*  http://mpeg-7.joanneum.at                                                */
/*                                                                           */
/*  JRS MPEG-7 Library is free software: you can redistribute it and/or      */
/*  modify it under the terms of the GNU Lesser General Public License       */
/*  (LGPL) as published by the Free Software Foundation, either version 3    */
/*  of the License, or (at your option) any later version.                   */
/*                                                                           */
/*  JRS MPEG-7 Library is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             */
/*  GNU Lesser General Public License for more details.                      */
/*                                                                           */
/*  You should have received a copy of the GNU Lesser General Public License */
/*  along with this library. If not, see <http://www.gnu.org/licenses/>.     */
/*****************************************************************************/



#include "Mp7JrsArchive.h"

#include <xercesc/framework/StdOutFormatTarget.hpp>
#include <xercesc/framework/LocalFileInputSource.hpp>
#include <xercesc/framework/LocalFileFormatTarget.hpp>
#include <xercesc/framework/MemBufInputSource.hpp>
#include <xercesc/framework/MemBufFormatTarget.hpp>
#include <xercesc/parsers/XercesDOMParser.hpp>
#include <xercesc/util/XMLString.hpp>
#include <xercesc/dom/DOM.hpp>
#include <xercesc/sax/ErrorHandler.hpp>
#include <xercesc/sax/SAXParseException.hpp>
#include "Mp7JrsMpeg7_LocalType.h" 
#include "Mp7JrsFactoryDefines.h"
#include "Mp7JrsNode.h"
#include "Mp7JrsPtr.h"
#include <assert.h>
#if defined (WIN32)
# include <typeinfo.h>
#else
# include <typeinfo>
#endif

class Mp7JrsErrorHandler : public ErrorHandler
{
public:

	Mp7JrsErrorHandler(){};
	~Mp7JrsErrorHandler(){};

	virtual void warning(const SAXParseException& exception)
	{
		XMLCh * msg = extractErrorMessage(exception); // FTT fixed mem leak
		Mp7JrsException e(MP7JRS_WARNING, msg);
		Mp7JrsUtil::deallocate(&msg);
		throw e;
	}

	virtual void error(const SAXParseException& exception)
	{
		XMLCh * msg = extractErrorMessage(exception); // FTT fixed mem leak
		Mp7JrsException e( MP7JRS_ERROR, msg);
		Mp7JrsUtil::deallocate(&msg);
		throw e;
	}

	virtual void fatalError(const SAXParseException& exception)
	{
		XMLCh * msg = extractErrorMessage(exception); // FTT fixed mem leak
		Mp7JrsException e( MP7JRS_FATAL_ERROR, msg);
		Mp7JrsUtil::deallocate(&msg);
		throw e;
	}

	void resetErrors(){};

private :
	XMLCh* extractErrorMessage(const SAXParseException& exception)
	{
		XMLCh * msg = XMLString::transcode("");
		Mp7JrsUtil::CatString(&msg, exception.getMessage());
		Mp7JrsUtil::CatString(&msg, X(". column: "));
		Mp7JrsUtil::CatString(&msg, Mp7JrsUtil::ConvertToXMLCh(exception.getColumnNumber()));
		Mp7JrsUtil::CatString(&msg, X("; line: "));
		Mp7JrsUtil::CatString(&msg, Mp7JrsUtil::ConvertToXMLCh(exception.getLineNumber()));
		Mp7JrsUtil::CatString(&msg, X("."));
		return msg;
	}

	/* Unimplemented constructors and operators */ 
	Mp7JrsErrorHandler(const ErrorHandler&);
	void operator=(const DOMErrorHandler&);
};

class Mp7JrsArchive::XercesWrapper 
{
public:
	XercesWrapper(InputSource &source, bool fullSchema, bool validate,
		const Mp7JrsSchemaLocationMap &locationMap);
	virtual ~XercesWrapper();
	DOMDocument * getDoc() const
	{ return doc_; }
	DOMElement * getRoot() const
	{ return root_; }
private:
	XercesWrapper();                // w/o implementation
	XercesWrapper &operator=(const XercesWrapper &rhs);     // w/o impl
	// Is this attribute necessary?
	DOMImplementation * impl_;
	XercesDOMParser * parser_;
	Mp7JrsErrorHandler * handler_;
	DOMDocument * doc_;
	DOMElement * root_;
};

Mp7JrsArchive::XercesWrapper::XercesWrapper(InputSource &source,
	bool fullSchema, bool validate, const Mp7JrsSchemaLocationMap &locationMap)
{
	bool useNS = locationMap.size() > 0; // This is new, we only use namespacehandling, if necessary

	// Is this necessary?
	impl_ =  DOMImplementationRegistry::getDOMImplementation(X("Core"));
	parser_ = new XercesDOMParser;
	
	// parser options
	parser_->setValidationSchemaFullChecking(fullSchema);
  // FTT This is for 2.7, 2.8 AND 3.0.1 (no deprecated setDoValidation())
  parser_->setValidationScheme(validate ? AbstractDOMParser::Val_Always : AbstractDOMParser::Val_Never);
	
	// So, if you don't define a default namespace (+prefix) and additional namespaces, then we skip namespacehandling
	parser_->setDoNamespaces(useNS);
	
	parser_->setDoSchema(validate);
	parser_->setExitOnFirstFatalError(true);
	XMLCh * txt = locationMap.ToText();      // FTT: setExternalSchemaLocation(const XMLCh const *)
	parser_->setExternalSchemaLocation(txt); // FTT: There is definitely something wrong with a language
	XMLString::release(&txt);                // FTT: where you can define "const char const *"
	handler_ = new Mp7JrsErrorHandler; 
	parser_->setErrorHandler(handler_);
	parser_->parse(source);
	doc_ = parser_->getDocument();
	root_ = doc_->getDocumentElement();
}

Mp7JrsArchive::XercesWrapper::~XercesWrapper()
{
	// This was factored out from _Deserialize; apparently, releasing
	// the document releases the parser as well. 
	// FTT: This is not the whole truth! Some parts of the parser seemed to be released, others not!
	delete handler_;
	// FTT: excluded this: doc_->release();
	delete parser_;
}

Mp7JrsArchive::Mp7JrsArchive(void)
{
	encoding = XMLString::transcode("UTF-16");
	
	destination = NULL;
	validate = false;
	fullSchema = false;
	writeXMLDecl = true;
	xmlDeclDefault = true;
	usePrefixForDefaultNS = false; // No prefix for default namespace
	prettyformat = false; // small xmlcode
	forceNamespaceDeclaration = true; // Declare ns in root not in each and every subnode
	standalone = true;
	version = XMLString::transcode("1.0");
	
	// Set default schema locations and namespaces
	// If no default namespace is given, skip it.
	
	// TODO: unify param order: uri,prefix vs prefix,uri
	_locationMap.SetDefaultNamespace( X("urn:mpeg:mpeg7:schema:2004"), X("mpeg7"));
	// There is a default schema location specified
	_locationMap.SetSchemaLocation(X("urn:mpeg:mpeg7:schema:2004"),X("mpeg7"),X("mpeg7-2004.xsd"));
	
	_locationMap.SetSchemaLocation(X("urn:mpeg:mpeg7:schema:2004"), X("mpeg7"), X("mpeg7-2004.xsd"));
	
}

Mp7JrsArchive::~Mp7JrsArchive(void)
{
	XMLString::release(&encoding);
	XMLString::release(&version);
}

void Mp7JrsArchive::ToFile(const XMLCh * path, Mp7JrsNodePtr &node, XMLCh* nodeName, bool useTypeAttr)
{
	destination = new LocalFileFormatTarget(path);
	// FTT This doesn't work for abstract or derived rootelements	
	//	bool writeHeader = node->GetClassId() == Mp7JrsFactory::GetTypeIndex(X("Mpeg7_LocalType"));
	bool writeHeader = ((Mp7JrsMpeg7_LocalPtr)node != 0);
	_Serialize(node, nodeName, useTypeAttr, writeHeader);
	
	delete destination;
	}

void Mp7JrsArchive::ToFile(const char* path, Mp7JrsNodePtr &node, XMLCh* nodeName, bool useTypeAttr) 
{
	if(!path || !node) {
		return;
	}
	ToFile(X(path),node,nodeName,useTypeAttr);
}

Mp7JrsNodePtr Mp7JrsArchive::FromFile(const XMLCh * path, Mp7JrsNodePtr &node)
{
	InputSource *source = new LocalFileInputSource(path);
	Mp7JrsNodePtr result = _Deserialize(node, *source);
	delete source;
	return result;
}

Mp7JrsNodePtr Mp7JrsArchive::FromFile(const XMLCh * path)
{
	Mp7JrsNodePtr empty = Mp7JrsNodePtr();
	return FromFile(path, empty);
}

Mp7JrsNodePtr Mp7JrsArchive::FromFile(const char * path, Mp7JrsNodePtr &node)
{
	return FromFile(X(path), node);
}

Mp7JrsNodePtr Mp7JrsArchive::FromFile(const char * path)
{
	Mp7JrsNodePtr empty = Mp7JrsNodePtr();
	return FromFile(X(path), empty);
}

//---
int Mp7JrsArchive::AddToCollectionFromFile(const XMLCh * path,
	Mp7JrsNodeCollectionPtr &collection, HowToAdd howtoadd, int where)
{
	InputSource *source = new LocalFileInputSource(path);
	int result =_DeserializeIntoCollection(*source, collection, howtoadd, where);
	delete source;
	return result;
}

int Mp7JrsArchive::AddToCollectionFromFile(const char * path,
  Mp7JrsNodeCollectionPtr &collection, HowToAdd howtoadd, int where)
{
	return AddToCollectionFromFile(X(path), collection, howtoadd, where);
}

int Mp7JrsArchive::AddToCollectionFromBuffer(const XMLCh * buffer,
  Mp7JrsNodeCollectionPtr &collection, HowToAdd howtoadd, int where)
{
	MemBufInputSource source((XMLByte*)buffer,XMLString::stringLen(buffer)*2,"id",false);
	source.setEncoding(X("UTF-16"));
	
	return _DeserializeIntoCollection(source, collection, howtoadd, where);
}

int Mp7JrsArchive::AddToCollectionFromBuffer(const char * buffer,
  Mp7JrsNodeCollectionPtr &collection, HowToAdd howtoadd, int where)
{
	return AddToCollectionFromBuffer(X(buffer), collection, howtoadd, where);
}

//---


XMLCh* Mp7JrsArchive::ToWBuffer(Mp7JrsNodePtr &node, XMLCh* nodeName, bool useTypeAttr) 
{
	destination = new MemBufFormatTarget();
	
	XMLCh * saveencoding = encoding;
	encoding = XMLString::transcode("UTF-16");
	
// FTT This doesn't work for abstract or derived rootelements	
//	bool writeHeader = node->GetClassId() == Mp7JrsFactory::GetTypeIndex(X("Mpeg7_LocalType"));
	bool writeHeader = ((Mp7JrsMpeg7_LocalPtr)node != 0);
	_Serialize(node, nodeName, useTypeAttr, writeHeader);
	
//	FTT don't do this, since people might use XMLString::release() -> on the wrong heap!
//       XMLCh* buf = new XMLCh[(((MemBufFormatTarget*)destination)->getLen()/2)+1];

// FTT NEW!!!! We consequently use the Xerces heap for this kind of stuff!!		
// So, call DeleteBuffer() after using it...
	XMLCh* buf = Mp7JrsUtil::allocate((long)((((MemBufFormatTarget*)destination)->getLen()/2)+1));
	XMLString::copyString (buf,(XMLCh*)(((MemBufFormatTarget*)destination)->getRawBuffer())); 
	
	delete destination;
	XMLString::release(&encoding);
	encoding = saveencoding;
	
	return buf;
}

char* Mp7JrsArchive::ToBuffer(Mp7JrsNodePtr &node, XMLCh* nodeName, bool useTypeAttr) 
{
	destination = new MemBufFormatTarget();
	
// FTT This doesn't work for abstract or derived rootelements	
//	bool writeHeader = node->GetClassId() == Mp7JrsFactory::GetTypeIndex(X("Mpeg7_LocalType"));
	bool writeHeader = ((Mp7JrsMpeg7_LocalPtr)node != 0);
	_Serialize(node, nodeName, useTypeAttr, writeHeader);

// FTT NEW!!!! We consequently use the Xerces heap for this kind of stuff!!		
// So, call DeleteBuffer() after using it...
	char * buf = NULL;
	if((((MemBufFormatTarget*)destination)->getLen()) > 0)
	{
		buf = XMLString::replicate((char*)(((MemBufFormatTarget*)destination)->getRawBuffer()));
	}
	delete destination;
	return buf;
}

Mp7JrsNodePtr Mp7JrsArchive::FromBuffer(const XMLCh * buf, Mp7JrsNodePtr &node)
{
	InputSource *source = new MemBufInputSource((XMLByte*)buf,XMLString::stringLen(buf)*2,"id",false);
	source->setEncoding(X("UTF-16"));
	
	Mp7JrsNodePtr result = _Deserialize(node, *source);
	
	delete source;
	return result;
}

Mp7JrsNodePtr Mp7JrsArchive::FromBuffer(const XMLCh * buf)
{
	Mp7JrsNodePtr empty = Mp7JrsNodePtr();
	return FromBuffer(buf, empty);
}

Mp7JrsNodePtr Mp7JrsArchive::FromBuffer(const char * buf, Mp7JrsNodePtr &node)
{
	InputSource *source = new MemBufInputSource((XMLByte*)buf,XMLString::stringLen(buf),"id",false);
	source->setEncoding(X("UTF-8"));
	
	Mp7JrsNodePtr result = _Deserialize(node, *source);
	
	delete source;
	return result;
}

Mp7JrsNodePtr Mp7JrsArchive::FromBuffer(const char * buf)
{
	Mp7JrsNodePtr empty = Mp7JrsNodePtr();
	return FromBuffer(buf, empty);
}

void Mp7JrsArchive::ToStdOut(Mp7JrsNodePtr &node, XMLCh* nodeName, bool useTypeAttr) 
{
	destination = new StdOutFormatTarget();

// FTT This doesn't work for abstract or derived rootelements	
//	bool writeHeader = node->GetClassId() == Mp7JrsFactory::GetTypeIndex(X("Mpeg7_LocalType"));
	bool writeHeader = ((Mp7JrsMpeg7_LocalPtr)node != 0);
	_Serialize(node, nodeName, useTypeAttr, writeHeader);
	
	delete destination;
}


void Mp7JrsArchive::_Serialize(Mp7JrsNodePtr &node, XMLCh* nodeName, bool useTypeAttr, bool writeHeader) 
{
#if XERCES_VERSION_MAJOR < 3
	DOMImplementation * impl =  DOMImplementationRegistry::getDOMImplementation(X("Core"));
#else
  DOMImplementation * impl = DOMImplementationRegistry::getDOMImplementation(X("LS"));
#endif // XERCES_VERSION_MAJOR < 3

	// FTT This doesn't work for abstract or derived rootelements	
	//	if (node->GetClassId() == Mp7JrsFactory::GetTypeIndex(X("Mpeg7_LocalType")))
	if((Mp7JrsMpeg7_LocalPtr)node != 0)
	{
		XMLString::release(&nodeName);
		nodeName = XMLString::transcode("Mpeg7");
	}
	else if (nodeName == NULL) 
	{
	  nodeName = XMLString::transcode("Mpeg7Fragment");
	}

	const XMLCh* defaultNS = 0;
	const XMLCh* prefix = 0;
	const XMLCh* ns = NULL;
	_locationMap.GetDefaultNamespace(&defaultNS, &prefix);
	
	DOMDocument * doc = impl->createDocument(defaultNS, nodeName, 0);
	
#if XERCES_VERSION_MAJOR < 3
	doc->setStandalone(standalone);
	doc->setEncoding(encoding);
	doc->setVersion(version);

	DOMWriter * theSerializer = ((DOMImplementationLS*)impl)->createDOMWriter();
#else	
	doc->setXmlStandalone(standalone);
//	doc->setXmlEncoding(encoding);
	doc->setXmlVersion(version);

  DOMLSSerializer * theSerializer = ((DOMImplementationLS*)impl)->createLSSerializer();
  DOMLSOutput * theOutput = ((DOMImplementationLS*)impl)->createLSOutput();
#endif // XERCES_VERSION_MAJOR < 3
  
#if defined(XML_WIN32) || defined(WIN32) // Xerces 3.0 doesn't know XML_WIN32
	theSerializer->setNewLine(X("\r\n"));
#elif defined(XML_MACOS) || defined(APPLE) || defined(__APPLE__)// Xerces 3.0 doesn't know XML_MACOS
	theSerializer->setNewLine(X("\n"));
#elif defined(XML_LINUX) || defined(LINUX) || defined(linux) // Xerces 3.0 doesn't know XML_LINUX
	theSerializer->setNewLine(X("\n"));
#else // XML_LINUX ...
#	error "TODO: Add platform specific headers for archive."
#endif
	
#if XERCES_VERSION_MAJOR < 3
	theSerializer->setFeature(XMLUni::fgDOMWRTFormatPrettyPrint, prettyformat);
#else
  if(theSerializer->getDomConfig()->canSetParameter(XMLUni::fgDOMWRTFormatPrettyPrint, true))
    theSerializer->getDomConfig()->setParameter(XMLUni::fgDOMWRTFormatPrettyPrint, prettyformat);
#endif // XERCES_VERSION_MAJOR < 3
	
	// baw 2004-02-04: save previous node name
	XMLCh* prevNodeName = XMLString::replicate(node->GetContentName());
	
	node->SetContentName(nodeName);
	node->UseTypeAttribute = useTypeAttr;
	
	DOMElement* nullNode = NULL;
	
	DOMElement * root = doc->getDocumentElement();
	
  // FTT For compatibility, we always write the namespace decls, 
  // but schema locations only if writeHeader is set.
	
	if ((writeHeader || forceNamespaceDeclaration) && _locationMap.size() > 0) {
		// Default namespace is already done (if exist)
		
    if(writeHeader || forceNamespaceDeclaration){
		// Add xmlns:prefix="URI" namespaces
		// Note, that if defaultnamespace has a prefix, but user sets
		// usePrefixForDefaultNS = false, then we skip adding the prefix for default ns.
		const XMLCh* prefix = NULL;
		const XMLCh* ns = NULL;
		bool ok = _locationMap.GetFirstNamespace(&prefix,&ns);
		while (ok) {
			if(prefix != 0 && XMLString::stringLen(prefix) > 0) {
			
				XMLCh * xmlns_prefix = 0;
				if(!XMLString::startsWith(prefix, X("xmlns:"))) // For compatibility
					xmlns_prefix = Mp7JrsUtil::CatString(X("xmlns:"), prefix);
				else 
					xmlns_prefix = XMLString::replicate(prefix);
					
				if(defaultNS != 0 && XMLString::compareString(ns, defaultNS) == 0) {
					if(usePrefixForDefaultNS) {
						root->setAttributeNS(X("http://www.w3.org/2000/xmlns/"), xmlns_prefix, ns); // Mind the trailing slash!
						const XMLCh* p = _locationMap.GetFirstSchemaPrefix(defaultNS);
						root->setPrefix(p);
					}
				}
				else {
					// Add xmlns:prefix="URI" attribute 
					root->setAttributeNS(X("http://www.w3.org/2000/xmlns/"), xmlns_prefix, ns); // Mind the trailing slash!
				}
				Mp7JrsUtil::deallocate(&xmlns_prefix);
			}
			if((prefix = _locationMap.GetNextSchemaPrefix(ns)) != 0)
				continue;
			
			ok = _locationMap.GetNextNamespace(&prefix,&ns);
		}
		// Add xsi namespace for xsi:type needed for schemalocation
		root->setAttributeNS(X("http://www.w3.org/2000/xmlns/"), X("xmlns:xsi"), X("http://www.w3.org/2001/XMLSchema-instance"));
    }
    		
		// Add schema location only to root node!
    if (writeHeader) {
      XMLCh* txt = _locationMap.ToText(); // TODO check this if more than one schema location
      root->setAttributeNS(X("http://www.w3.org/2001/XMLSchema-instance"), X("xsi:schemaLocation")
		    , txt);
    }
	}
	
	node->Serialize(doc, nullNode, &root);
	
	
	// baw 09 04 2003 
	// Write XMLDecl if desired
	// Has to be done by hand, as this is not supported in DOM 2 impl
	bool doWriteXMLDecl = writeXMLDecl;
	// If default value is set, determine if whole document or fragment is to be written
	if (xmlDeclDefault) {
		// FTT This doesn't work for abstract or derived rootelements	
		// if (node->GetClassId() == Mp7JrsFactory::GetTypeIndex(X("Mpeg7_LocalType"))) 
		if((Mp7JrsMpeg7_LocalPtr)node != 0)
			doWriteXMLDecl = true;
		else 
			doWriteXMLDecl = false;
	}
	
	if (doWriteXMLDecl) {
		// create the string, fixed len = 45
		XMLSize_t /*unsigned int*/ len = 45;
		// add length of other items to be written
		len += XMLString::stringLen(version);
		len += XMLString::stringLen(encoding);
		if (standalone) len+=3; 
		else len+=3;
		
		// build string
		XMLCh* xmlDeclStr = new XMLCh[len+1];
		xmlDeclStr[0] = '\0';
		
		XMLString::catString(xmlDeclStr,X("<?xml version=\""));
		XMLString::catString(xmlDeclStr,version);
		XMLString::catString(xmlDeclStr,X("\" encoding=\""));
		XMLString::catString(xmlDeclStr,encoding);
		XMLString::catString(xmlDeclStr,X("\" standalone=\""));
		if (standalone) XMLString::catString(xmlDeclStr,X("yes"));
		else XMLString::catString(xmlDeclStr,X("no"));
		XMLString::catString(xmlDeclStr,X("\" ?>"));
		
		// write to formatter
		XMLFormatter formatter(encoding,destination);
		// 8 bit                
		if ( (XMLString::compareString(encoding,X("UTF-8"))==0)
			||(XMLString::compareString(encoding,X("ISO-8859-1"))==0)) 
		{
			char* xmlDeclStrA = XMLString::transcode(xmlDeclStr);
			destination->writeChars((const XMLByte*) xmlDeclStrA, len, &formatter); 
			XMLString::release(&xmlDeclStrA);
		}
		// 16 bit
		else {
			destination->writeChars((const XMLByte*) xmlDeclStr, len * 2, &formatter); 
		}
	          
		delete [] xmlDeclStr;
	}
	
#if XERCES_VERSION_MAJOR < 3
	theSerializer->writeNode(destination, *root);
#else
  theOutput->setEncoding(encoding);
	theOutput->setByteStream(destination);
	theSerializer->write(root, theOutput);
	theOutput->release();
#endif // XERCES_VERSION_MAJOR < 3	

	theSerializer->release();
	doc->release();
	
	// set content name back to previous value
	node->SetContentName(prevNodeName);
}

int Mp7JrsArchive::_DeserializeIntoCollection(InputSource &source,
	Mp7JrsNodeCollectionPtr &collection, HowToAdd howtoadd, int where)
{
	// don't set encoding here. source should know encoding already.
	
	XercesWrapper wrapper(source, fullSchema, validate, _locationMap);
	if (!wrapper.getRoot()) return -1;
	if (!collection) return -1;
	
	unsigned int old_size = collection->size();
	
	DOMElement * dummy = NULL;
	collection->Deserialize(wrapper.getDoc(), wrapper.getRoot(), & dummy);
	unsigned int new_size = collection->size();
	
	if (old_size == new_size) return -1;
	
	Mp7JrsNodePtr new_node;
	
	switch (howtoadd) {
	case APPEND:
		// Nothing to do.
		return collection->size() - 1;
	case INSERT:
		assert(where >= 0 && where < (int)collection->size());
		if (!(where >= 0 && where < (int)collection->size())) return -1;
		new_node = collection->elementAt(new_size - 1);
		collection->removeElementAt(new_size - 1);
		collection->insertElementAt(new_node, where);
		return where;
	case REPLACE:
		assert(where >= 0 && where < (int)collection->size());
		if (!(where >= 0 && where < (int)collection->size())) return -1;
		new_node = collection->elementAt(new_size - 1);
		collection->removeElementAt(new_size - 1);
		collection->setElementAt(new_node, where);
		return where;
	default:
		return -1;
	}
}

Mp7JrsNodePtr Mp7JrsArchive::_Deserialize(Mp7JrsNodePtr &parentnode, InputSource &source)
{
	XercesWrapper wrapper(source, fullSchema, validate, _locationMap);
	bool isFragment = false;
	Mp7JrsNodePtr rootNode;

  if(!wrapper.getRoot()) { // Document is empty
    return Mp7JrsNodePtr(); // Bail out
  }
  
  if(parentnode == NULL) // Create decent root node
  {  
    if (Mp7JrsUtil::HasNodeName(wrapper.getRoot(), X("Mpeg7"), X("urn:mpeg:mpeg7:schema:2004")))
    {
 		  rootNode = CreateMpeg7_LocalType;
    } 
    else
    {
      XMLCh * msg = Mp7JrsUtil::CatString(X("Mp7JrsArchive::_Deserialize(): XML root nodes other than "), X("Mpeg7"));
      Mp7JrsUtil::CatString(&msg, X(" need an xsi:type attribute to be deserialised correctly."));
      Mp7JrsException e(MP7JRS_ERROR, msg);
      Mp7JrsUtil::deallocate(&msg);
      throw e;
    }
  }
  else // Check if node matches doc root type
  {		
    const XMLCh* name = wrapper.getRoot()->getNodeName();
//    const XMLCh* xsitype = wrapper.getRoot()->getAttribute(X("xsi:type"); // FTT for now xsi prefix is assumed
    NamespaceHelper result;
    Mp7JrsUtil::GetXsiType(wrapper.getRoot(), wrapper.getDoc(), &result);
    // Create child node based on the document's root node's type information
    XMLCh * xsitype = Mp7JrsUtil::QualifiedName(&result);
    
    if(!xsitype || XMLString::stringLen(xsitype) == 0) // No hint given, so is it root?
    {
      if(xsitype) Mp7JrsUtil::deallocate(&xsitype);
      if (Mp7JrsUtil::HasNodeName(wrapper.getRoot(), X("Mpeg7"), X("urn:mpeg:mpeg7:schema:2004")))
      {
 	  	  rootNode = parentnode;
 	    }
 	    else
 	    {
        rootNode = parentnode->CreateChild(name, xsitype);
        if(!rootNode) {
          XMLCh * msg = Mp7JrsUtil::CatString(X("Mp7JrsArchive::_Deserialize(): "), X("XML root node name "));
          Mp7JrsUtil::CatString(&msg, name);
          Mp7JrsUtil::CatString(&msg, X(" is not matching or xsi:type attribute is missing."));
          Mp7JrsException e(MP7JRS_ERROR, msg);
          Mp7JrsUtil::deallocate(&msg);
          throw e;
        }         
        isFragment = true; // Ok, this is a fragment
 	    }
    } else { // We have an xsi:type hint
      // Create child node based on the document's root node's type information
      isFragment = true; // Ok, this is a fragment, so use the node as parent
      rootNode = parentnode->CreateChild(name, xsitype);
      Mp7JrsUtil::deallocate(&xsitype);
    }
  }    
#if 0      
    
    else {
      // Create child node based on the document's root node's type information
      rootNode = node->CreateChild(type, xsitype);
    }
  }
 else
	  	throw Mp7JrsException (MP7JRS_ERROR, "Sorry, root element is abstract and needs a decent root node to start deserialization.");

				if (!Mp7JrsUtil::HasNodeName(wrapper.getRoot(), X("Mpeg7"), X("urn:mpeg:mpeg7:schema:2004")))
				{
					// No valid root node in document.
					XMLCh * msg = Mp7JrsUtil::CatString(X("Mp7JrsArchive::_Deserialize(): Input source is not a valid MPEG-7 document. Root node "), X("Mpeg7"));
					Mp7JrsUtil::CatString(&msg, X(" node or proper fragment pointer is expected."));
					Mp7JrsException e(MP7JRS_ERROR, msg);
					Mp7JrsUtil::deallocate(&msg);
					throw e;
				}
  		  rootNode = CreateMpeg7_LocalType;

	  } else { // Given node and root
			if (wrapper.getRoot()) { // Check if document root exists
				if (!Mp7JrsUtil::HasNodeName(wrapper.getRoot(), X("Mpeg7"), X("urn:mpeg:mpeg7:schema:2004"))
				{
					// No valid root node in document.
					XMLCh * msg = Mp7JrsUtil::CatString(X("Mp7JrsArchive::_Deserialize: Input source is no valid MPEG-7 document. Root node "), X("Mpeg7"));
					Mp7JrsUtil::CatString(&msg, X(" is expected."));
					Mp7JrsException e(MP7JRS_ERROR, msg);
					Mp7JrsUtil::deallocate(&msg);
					throw e;
				}
			}
			rootNode = node; // Ok this is a root node
		} else {
		
			isFragment = true; // Ok, this is a fragment, so use the node as parent
			const XMLCh * name = wrapper.getRoot()->getNodeName();
			// const XMLCh* xsitype = wrapper.getRoot()->getAttribute(XMLString::transcode("xsi:type"));
			NamespaceHelper result;
			Mp7JrsUtil::GetXsiType(wrapper.getRoot(), wrapper.getDoc(), &result);
			// Create child node based on the document's root node's type information
			XMLCh * xsitype = Mp7JrsUtil::QualifiedName(&result);
			rootNode = node->CreateChild(name, xsitype);
			Mp7JrsUtil::deallocate(&xsitype);
		}
	}
#endif

	if (rootNode != NULL) {
		DOMElement * dummy = NULL;
		rootNode->Deserialize(wrapper.getDoc(), wrapper.getRoot(), & dummy);
		// baw 2004-11-15: set content name of root node only if we are not parsing into a parent node
		if (!isFragment) 
			rootNode->SetContentName(XMLString::replicate(wrapper.getRoot()->getNodeName()));   
	}
	return rootNode;

#if 0	
/////////////////////////////
  if (wrapper.getRoot()) { // ok, there is a document
    if (node == NULL) { // deserialize whole mpeg-7 document
      if (XERCES_CPP_NAMESPACE_QUALIFIER XMLString::compareString(
	  wrapper.getRoot()->getNodeName(), X("Mpeg7")) == 0) {
	rootNode = CreateMpeg7_LocalType;
      } else {
	// no valid root node in document.
	throw Mp7JrsException(MP7JRS_WARNING, "Mp7JrsArchive::_Deserialize: Input source is no mpeg-7 document. Root node not found.");
      }
    } else { // parse fragment into node
      const XMLCh* type = wrapper.getRoot()->getNodeName();
      const XMLCh* xsitype = wrapper.getRoot()->getAttribute(XERCES_CPP_NAMESPACE_QUALIFIER XMLString::transcode("xsi:type"));
      // create child node based on the document's root node's type information
      rootNode = node->CreateChild(type, xsitype);
    }
    
    if (rootNode != NULL) {
      XERCES_CPP_NAMESPACE_QUALIFIER DOMElement * dummy = NULL;
      rootNode->Deserialize(wrapper.getDoc(), wrapper.getRoot(), & dummy);
      // baw 2004-11-15: set content name of root node only if we are not parsing into a parent node
      if (node == NULL) rootNode->SetContentName(XMLString::replicate(wrapper.getRoot()->getNodeName()));   
    }
  }
////////////////////////////	
#endif 

  return rootNode;
}

void Mp7JrsArchive::SetEncoding(const XMLCh * encoding) 
{
	XMLString::release(&(this->encoding));
	this->encoding = XMLString::replicate(encoding);
}

char * Mp7JrsArchive::GetEncoding() 
{
	return XMLString::transcode(encoding);
}


void Mp7JrsArchive::SetSchemaLocation(const XMLCh* namespaceURI, const XMLCh* prefix, const XMLCh * schemaLocation) 
{
	_locationMap.SetSchemaLocation(namespaceURI,prefix,schemaLocation);
}

char * Mp7JrsArchive::GetSchemaLocation(const char* namespaceURI) 
{
	return XMLString::transcode(_locationMap.GetSchemaLocation(X(namespaceURI)));
}

void Mp7JrsArchive::SetDefaultNamespace(const XMLCh* namespaceURI, const XMLCh* prefix) 
{
	_locationMap.SetDefaultNamespace(namespaceURI, prefix);
}

bool Mp7JrsArchive::GetDefaultNamespace(const XMLCh* * namespaceURI, const XMLCh* * prefix) 
{
	return _locationMap.GetDefaultNamespace(namespaceURI, prefix);
}


void Mp7JrsArchive::DeleteBuffer(char** buffer) 
{
	// This is now on Xerces heap, TODO add Util::allocate() /deallocate() for 8 bit strings as well
	XMLString::release(buffer);
}

void Mp7JrsArchive::DeleteBuffer(XMLCh** buffer) 
{
	// This is now on Xerces heap
	Mp7JrsUtil::deallocate(buffer);
}

