
/*****************************************************************************/
/* Copyright (C) 2003-2010, JOANNEUM RESEARCH Forschungs Ges.m.b.H.          */
/* Steyrergasse 17, A-8010 Graz, Austria.                                    */
/*                                                                           */
/*  This file is part of the JRS MPEG-7 Library.                             */
/*  http://mpeg-7.joanneum.at                                                */
/*                                                                           */
/*  JRS MPEG-7 Library is free software: you can redistribute it and/or      */
/*  modify it under the terms of the GNU Lesser General Public License       */
/*  (LGPL) as published by the Free Software Foundation, either version 3    */
/*  of the License, or (at your option) any later version.                   */
/*                                                                           */
/*  JRS MPEG-7 Library is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             */
/*  GNU Lesser General Public License for more details.                      */
/*                                                                           */
/*  You should have received a copy of the GNU Lesser General Public License */
/*  along with this library. If not, see <http://www.gnu.org/licenses/>.     */
/*****************************************************************************/


// Based on ClientManager_cpp.template

#include "Mp7JrsClientManager.h"

#include "Mp7JrsNodeLockNotification.h"
#include "Mp7JrsNodeUpdateNotification.h"

Mp7JrsClientID Mp7JrsClientManager::_maxID = 0;
ClientMap Mp7JrsClientManager::_clientMap(20);

LockQueue Mp7JrsClientManager::_lockQueue(20);

Mp7JrsListenerList Mp7JrsClientManager::_lockRequested;
Mp7JrsListenerList Mp7JrsClientManager::_lockGranted;
Mp7JrsListenerList Mp7JrsClientManager::_unlocked;
Mp7JrsListenerList Mp7JrsClientManager::_added;
Mp7JrsListenerList Mp7JrsClientManager::_inserted;
Mp7JrsListenerList Mp7JrsClientManager::_replaced;

bool Mp7JrsClientManager::_lockingEnabled = false;
bool Mp7JrsClientManager::_notificationsEnabled = false;

Mp7JrsListenerList::Mp7JrsListenerList() :
	Mp7JrsValueVectorOf<Mp7JrsClientRNodePair>(20)
{}

Mp7JrsClientList Mp7JrsListenerList::GetClientsToBeNotified(Mp7JrsNodePtr node) {
	Mp7JrsClientList toBeNotified(20);

	int nElems = size();
	int i;
	for (i=0;i<nElems;i++) {
		Mp7JrsClientRNodePair& crnPair = elementAt(i);
		// check if node itself is affected
		bool notify = false;
		if (crnPair.getValue().node==node) notify = true;
		// check if parent is affected if registration is recursive
		if ((!notify) && (crnPair.getValue().recursive)) {
			Mp7JrsNodePtr tmpNode = node->GetParent();
			while (tmpNode!=NULL) {
				if (tmpNode==crnPair.getValue().node) {
					notify = true;
					break;
				}
				else tmpNode = tmpNode->GetParent();
			}
		}

		// add to list
		if (notify) toBeNotified.addElement(crnPair.getKey());
	}
	
	return toBeNotified;
}


Mp7JrsClientID Mp7JrsClientManager::GetNewID(Mp7JrsNotificationListener* listener) {
	Mp7JrsClientID newID = ++_maxID;

	ClientListenerPair clPair(newID,listener);
	Mp7JrsClientManager::_clientMap.addElement(clPair);

	return newID;
}

Mp7JrsNotificationListener* Mp7JrsClientManager::GetListener(Mp7JrsClientID id) {
	int nElems = Mp7JrsClientManager::_clientMap.size();
	int i;
	for (i=0;i<nElems;i++) {
		ClientListenerPair& clPair = _clientMap.elementAt(i);
		if (clPair.getKey()==id) return clPair.getValue();
	}
	return NULL;
}


void Mp7JrsClientManager::Unregister(Mp7JrsClientID id) {
	int nElems = Mp7JrsClientManager::_clientMap.size();
	int i;
	for (i=0;i<nElems;i++) {
		ClientListenerPair& clPair = _clientMap.elementAt(i);
		if (clPair.getKey()==id) {
			_clientMap.removeElementAt(i);
			return;
		}
	}	
}

void Mp7JrsClientManager::QueueForLock(Mp7JrsNodePtr sender, Mp7JrsClientID client, bool /*forAdd*/) {
	// check if already queued
	int nElems = Mp7JrsClientManager::_lockQueue.size();
	int i;
	bool found = false;

	for (i=0;i<nElems;i++) {
		ClientNodePair& cnPair = _lockQueue.elementAt(i);
		if (cnPair.getKey()==client) {
			found = true;
			break;
		}
	}	

	// if not found, queue
	if (!found) {
		ClientNodePair cnPair(client,sender);
		_lockQueue.addElement(cnPair);
	}

}

bool Mp7JrsClientManager::SendLockRequestNotification(Mp7JrsNodePtr sender, Mp7JrsClientID client, bool forAdd, bool notify) {
	if (!_notificationsEnabled) return true;
	
	// bool queued = false;
	
	// create and send notification about request

	Mp7JrsNodeLockNotification lockNotif;
	lockNotif.action = MP7JRS_NA_LOCK_REQUESTED;
	lockNotif.clientId = client;
	lockNotif.node = sender;
	lockNotif.lock4add = forAdd;
	lockNotif.notifyLater = notify;
	lockNotif.result = false;

	bool grantLock = true;

	Mp7JrsClientList toBeNotified = _lockRequested.GetClientsToBeNotified(sender);
	unsigned int i;
	for (i=0;i<toBeNotified.size();i++) {
		Mp7JrsNotificationListener* listener = GetListener(toBeNotified.elementAt(i));
		if (listener) {
			listener->Notify(lockNotif);
			if (!lockNotif.result) grantLock = false;
		}
	}
	
	return grantLock;
}

void Mp7JrsClientManager::SendLockNotification(Mp7JrsNodePtr sender, Mp7JrsClientID client, bool forAdd) {
	if (!_notificationsEnabled) return;

	Mp7JrsNodeLockNotification lockNotif;
	lockNotif.action = MP7JRS_NA_LOCK_GRANTED;
	lockNotif.clientId = client;
	lockNotif.node = sender;
	lockNotif.lock4add = forAdd;

	Mp7JrsClientList toBeNotified = _lockGranted.GetClientsToBeNotified(sender);
	unsigned int i;
	for (i=0;i<toBeNotified.size();i++) {
		Mp7JrsNotificationListener* listener = GetListener(toBeNotified.elementAt(i));
		if (listener) {
			listener->Notify(lockNotif);
		}
	}
}


void Mp7JrsClientManager::SendUnlockNotification(Mp7JrsNodePtr sender) {
	if (!_notificationsEnabled) return;

	// first grant lock to next in queue
	if (_lockQueue.size()>0) {
		// check if the client is already registered for notification,
		// if not, register it temporarily
		bool registerTemp = true;

		int nElems = _lockGranted.size();
		int i;
		for (i=0;i<nElems;i++) {
			Mp7JrsClientRNodePair& crnPair = _lockGranted.elementAt(i);
			if (crnPair.getKey()==_lockQueue.elementAt(0).getKey()) {
				registerTemp = false;
				break;
			}	
		}

		if (registerTemp) Register(_lockQueue.elementAt(0).getKey(),MP7JRS_NA_LOCK_GRANTED,_lockQueue.elementAt(0).getValue(),false);

		// request the lock at the node
		_lockQueue.elementAt(0).getValue()->Lock(_lockQueue.elementAt(0).getKey(),false);
		
		if (registerTemp) Unregister(_lockQueue.elementAt(0).getKey(),MP7JRS_NA_LOCK_GRANTED,_lockQueue.elementAt(0).getValue());
		
		// remove from queue
		_lockQueue.removeElementAt(0);
	}

	// then inform others about unlock
	Mp7JrsNodeLockNotification lockNotif;
	lockNotif.action = MP7JRS_NA_UNLOCK;
	lockNotif.node = sender;

	Mp7JrsClientList toBeNotified = _unlocked.GetClientsToBeNotified(sender);
	unsigned int i;
	for (i=0;i<toBeNotified.size();i++) {
		Mp7JrsNotificationListener* listener = GetListener(toBeNotified.elementAt(i));
		if (listener) {
			listener->Notify(lockNotif);
		}
	}

}
	
void Mp7JrsClientManager::SendUpdateNotification(Mp7JrsNodePtr sender, Mp7JrsNotificationAction action, Mp7JrsClientID client, Mp7JrsNodePtr second, int index) {
	if (!_notificationsEnabled) return;
	
	Mp7JrsNodeUpdateNotification updateNotif;
	updateNotif.action = action;
	updateNotif.clientId = client;
	updateNotif.node = sender;
	updateNotif.newNode = second;
	updateNotif.index = index;

	Mp7JrsListenerList& ll = _added;
	
	switch (action) {
	case MP7JRS_NA_NODE_ADDED: ll = Mp7JrsClientManager::_added; break;
	case MP7JRS_NA_NODE_INSERTED: ll = Mp7JrsClientManager::_inserted; break;
	case MP7JRS_NA_NODE_UPDATED: ll = Mp7JrsClientManager::_replaced; break;
	default:;
	};


	Mp7JrsClientList toBeNotified = ll.GetClientsToBeNotified(sender);
	unsigned int i;
	for (i=0;i<toBeNotified.size();i++) {
		Mp7JrsNotificationListener* listener = GetListener(toBeNotified.elementAt(i));
		if (listener) {
			listener->Notify(updateNotif);
		}
	}
}


void Mp7JrsClientManager::Register(Mp7JrsClientID id, Mp7JrsNotificationAction action, Mp7JrsNodePtr node, bool recursive) {
	Mp7JrsListenerList& ll = _lockRequested;
	
	switch (action) {
	case MP7JRS_NA_LOCK_REQUESTED: ll = Mp7JrsClientManager::_lockRequested; break;
	case MP7JRS_NA_LOCK_GRANTED: ll = Mp7JrsClientManager::_lockGranted; break;
	case MP7JRS_NA_UNLOCK: ll = Mp7JrsClientManager::_unlocked; break;
	case MP7JRS_NA_NODE_ADDED: ll = Mp7JrsClientManager::_added; break;
	case MP7JRS_NA_NODE_INSERTED: ll = Mp7JrsClientManager::_inserted; break;
	case MP7JRS_NA_NODE_UPDATED: ll = Mp7JrsClientManager::_replaced; break;
	};

	Subtree st;
	st.node = node;
	st.recursive = recursive;
	Mp7JrsClientRNodePair crnPair(id,st);

	ll.addElement(crnPair);
}
	
void Mp7JrsClientManager::Unregister(Mp7JrsClientID id, Mp7JrsNotificationAction action, Mp7JrsNodePtr node) {
	Mp7JrsListenerList& ll = _lockRequested;
	
	switch (action) {
	case MP7JRS_NA_LOCK_REQUESTED: ll = Mp7JrsClientManager::_lockRequested; break;
	case MP7JRS_NA_LOCK_GRANTED: ll = Mp7JrsClientManager::_lockGranted; break;
	case MP7JRS_NA_UNLOCK: ll = Mp7JrsClientManager::_unlocked; break;
	case MP7JRS_NA_NODE_ADDED: ll = Mp7JrsClientManager::_added; break;
	case MP7JRS_NA_NODE_INSERTED: ll = Mp7JrsClientManager::_inserted; break;
	case MP7JRS_NA_NODE_UPDATED: ll = Mp7JrsClientManager::_replaced; break;
	};

	int nElems = ll.size();
	int i;
	for (i=0;i<nElems;i++) {
		Mp7JrsClientRNodePair& crnPair = ll.elementAt(i);
		if ((crnPair.getKey()==id) && (crnPair.getValue().node==node)) {
			ll.removeElementAt(i);
			return;
		}
	}

}

// FTT moved from headerfile to allows extension DLLs

void Mp7JrsClientManager::SetNotificationsEnabled(bool enabled) { _notificationsEnabled = enabled; }
void Mp7JrsClientManager::SetLockingEnabled(bool enabled) { _lockingEnabled = enabled; }
bool Mp7JrsClientManager::IsNotificationsEnabled() { return _notificationsEnabled; };
bool Mp7JrsClientManager::IsLockingEnabled() { return _lockingEnabled; };

